package v7.name.panitz.wip;

import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.HashMap;
import java.util.Map;

import v7.name.panitz.wip.Tree.ConDef;
import v7.name.panitz.wip.Tree.FieldSelect;
import v7.name.panitz.wip.Tree.FunCall;
import v7.name.panitz.wip.Tree.FunDef;
import v7.name.panitz.wip.Tree.IfExpr;
import v7.name.panitz.wip.Tree.IntLiteral;
import v7.name.panitz.wip.Tree.OpExpr;
import v7.name.panitz.wip.Tree.Param;
import v7.name.panitz.wip.Tree.Program;
import v7.name.panitz.wip.Tree.Variable;

public class GenCode implements Visitor<Void> {
	Writer w;
	Map<String, Integer> vars;

	int label = 0;
	private FunDef currentFun;
	String nextLabel() {
		return ".L"+(label++);
	}
	
	public GenCode() {
		try {
			w = new FileWriter("main.s");
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	void finish() {
		try {
			w.write("  ret\n");
			w.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} 
		
	}
	
	@Override
	public Void visit(IntLiteral il) {
		// TODO Auto-generated method stub
		try {
			w.write("  movq $"+il.n+", %rax\n");
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return null;
	}

	@Override
	public Void visit(OpExpr oe)  {
		try {
			oe.rhs.welcome(this);
			w.write("  pushq %rax\n");
			oe.lhs.welcome(this);
			w.write("  popq %rbx\n");
			switch (oe.op) {
			case mult: {
				w.write("  imulq %rbx, %rax\n"); 
				break;
			}
			case sub: {
				w.write("  subq %rbx, %rax\n"); 
				break;
			}
			default:
				w.write("  addq %rbx, %rax\n"); 
			}
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public Void visit(ConDef conDef) {
		try {
			w.append("\n  .globl "+conDef.name+"\n");
			w.append("  .type "+conDef.name+", @function\n\n");

			w.append(conDef.name+":\n");
			w.append("  pushq %rbp\n");
			w.append("  movq %rsp, %rbp\n");

			w.append("  movl	$"+(8*conDef.args.size())+", %edi\n");
			w.append("  call	malloc@PLT\n");

			var i = conDef.args.size()+1;
			var j = 0;
			for (var a:conDef.args) {
				//vars.put(a.name, (i--)*8);
				w.append("  movq	"+(i--)*8+"(%rbp), %rdx\n");
				w.append("  movq	%rdx, "+j+"(%rax)\n");
				j+=8;
			}
			
			w.append("  movq %rbp, %rsp\n");
			w.append("  popq %rbp\n");
			w.append("  ret\n\n");
			
		}catch (Exception e) {
			// TODO: handle exception
		}
		return null;
	}

	@Override
	public Void visit(FunDef funDef) {
		currentFun = funDef;
		try {
			w.append("\n  .globl "+funDef.name+"\n");
			w.append("  .type "+funDef.name+", @function\n\n");
			
			w.append(funDef.name+":\n");
			
			w.append("  pushq %rbp\n");
			w.append("  movq %rsp, %rbp\n");

			w.append(".L"+funDef.name+":\n");

			
			vars = new HashMap<>();
			var i = funDef.args.size()+1;
			for (var a:funDef.args)vars.put(a.name, (i--)*8);
			
			funDef.body.welcome(this);
			
			w.append("  movq %rbp, %rsp\n");
			w.append("  popq %rbp\n");
			w.append("  ret\n\n");
			
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		return null;
	}
	@Override
	public Void visit(Program program) {
		program.cons.forEach(con->con.welcome(this));
		program.funs.forEach(fun->fun.welcome(this));

//		program.main.welcome(this);
		return null;
	}
	@Override
	public Void visit(Variable variable) {
		try {
			w.append("  movq "+vars.get(variable.name)+"(%rbp), %rax\n");
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return null;
	}
	@Override
	public Void visit(FunCall funCall) {
		funCall.params.forEach((param)->{
			param.welcome(this);
			try {
				w.append("  pushq %rax\n");
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		});
		try {
			if (funCall.isTailRecursive){
				for (int i = 0;i<funCall.params.size();i++) {
					w.append("  popq %rax\n");
					w.append("  movq %rax, "+((2+i)*8)+"(%rbp)\n");
				}
				w.append("  jmp .L"+funCall.name+"\n");				
			}else {
				w.append("  call "+funCall.name+"\n");
				w.append("  addq $"+(8*funCall.params.size())+", %rsp\n");
			}
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return null;
	}
	@Override
	public Void visit(Param param) {
		// TODO Auto-generated method stub
		return null;
	}
	@Override
	public Void visit(FieldSelect fieldSelect) {
		fieldSelect.o.welcome(this);
		
		try {
			w.append("  movq	"+8*fieldSelect.nr+"(%rax), %rax\n");
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		return null;
	}
	@Override
	public Void visit(IfExpr ifExpr) {
		ifExpr.cond.welcome(this);
		var endLabel = nextLabel();
		var falseLabel = nextLabel();
		try {
			w.append("  cmpq	$0, %rax\n");
			w.append("  je	"+falseLabel+"\n");
			ifExpr.a1.welcome(this);
			w.append("  jmp	"+endLabel+"\n");
			w.append(falseLabel+":\n");
			ifExpr.a2.welcome(this);
			w.append(endLabel+":\n");

		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		return null;
	}

}
