package v7.name.panitz.wip;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import v7.name.panitz.wip.Tree.ConDef;
import v7.name.panitz.wip.Tree.FieldSelect;
import v7.name.panitz.wip.Tree.FunCall;
import v7.name.panitz.wip.Tree.FunDef;
import v7.name.panitz.wip.Tree.IfExpr;
import v7.name.panitz.wip.Tree.IntLiteral;
import v7.name.panitz.wip.Tree.OpExpr;
import v7.name.panitz.wip.Tree.Param;
import v7.name.panitz.wip.Tree.Program;
import v7.name.panitz.wip.Tree.Variable;

public class TypeCheck implements Visitor<String>{
	Map<String, Tree.FunDef> funs;
	Map<String, Tree.ConDef> cons;
	Map<String, String> vars;

	@Override
	public String visit(IntLiteral il) {
		il.type="int";
		return "int";
	}

	@Override
	public String visit(OpExpr oe) {
		var lt = oe.lhs.welcome(this);
		var rt = oe.rhs.welcome(this);
		if (!lt.equals("int")) throw new RuntimeException("kein int in Operand. Typ des Operanden: "+lt);
		if (!rt.equals("int")) throw new RuntimeException("kein int in Operand. Typ des Operanden: "+rt);
		oe.type="int";
		return "int";
	}

	@Override
	public String visit(FunDef funDef) {
		vars = new HashMap<>();

		funDef.args.forEach(a -> {
			var t = a.typ;
			vars.put(a.name, t);
			if (!t.equals("int")&&cons.get(t)==null) {
				throw new RuntimeException("unknown type: "+t);
			}
		});
		var t = funDef.body.welcome(this);
		if (!t.equals(funDef.type)) {
			throw new RuntimeException("wrong result type for function "+funDef.name+". Declared type: "+funDef.type+" Actual type: "+t);
		}
		return t;
	}

	@Override
	public String visit(ConDef conDef) {
		conDef.args.forEach(a -> {
			var t = a.typ;
			if (!t.equals("int")&&cons.get(t)==null) {
				throw new RuntimeException("unknown type: "+t);
			}
		});
		conDef.type = conDef.name;
		return conDef.type;
	}

	@Override
	public String visit(Param param) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public String visit(Program program) {
		funs = new HashMap<>();
		program.funs.forEach(fun->funs.put(fun.name, fun));
		cons = new HashMap<>();
		program.cons.forEach(con->cons.put(con.name, con));
		program.cons.forEach(con->con.welcome(this));
		program.funs.forEach(fun->fun.welcome(this));
		
	//	program.type = program.main.welcome(this);
		return "int"; //program.type;
	}

	@Override
	public String visit(Variable variable) {
		var t = vars.get(variable.name);
		if (t==null) {
			throw new RuntimeException("Unkown variable: "+variable.name);
		}
		variable.type = t;
		return t;
	}

	@Override
	public String visit(FunCall funCall) {
		var fun = funs.get(funCall.name);
		if (fun!=null) {
			var args = fun.args;
			checkParameters(funCall, args);
			return funCall.type = fun.type;
		}else {
			var con = cons.get(funCall.name);
			if (con!=null) {
				var args = con.args;
				checkParameters(funCall, args);
			
				return funCall.type = con.type;		
			}
		}
		throw new RuntimeException("unknown function "+funCall.name);
	}

	private void checkParameters(FunCall funCall, List<Param> args) {
		if (args.size()!=funCall.params.size()) {
			throw new RuntimeException("Wrong number of arguments for function "+funCall.name);
		}	
		var as = args.iterator();
		var params = funCall.params.iterator();
		while(as.hasNext()) {
			var arg = as.next();
			var param = params.next();
			var pt = param.welcome(this);
			if (!arg.typ.equals(pt)) {
				throw new RuntimeException("Wrong parameter type for function "+funCall.name);
			}
		}
	}

	@Override
	public String visit(FieldSelect fieldSelect) {
		var ot= fieldSelect.o.welcome(this);
		var con = cons.get(ot);
		if (con==null) throw new RuntimeException("field selection on  non struct type: "+ot);
		
		if (!con.args.stream().anyMatch(arg->arg.name.equals(fieldSelect.field))) {
			throw new RuntimeException("unknown field selection on struct type: "+ot+ " field "+fieldSelect.field);
		}
		var i = 0;
		for (var arg:con.args) {
			if (arg.name.equals(fieldSelect.field)) {
				fieldSelect.nr = i;
				return fieldSelect.type=arg.typ;
			}
			i++;
		}
		return null;
	}

	@Override
	public String visit(IfExpr ifExpr) {
		ifExpr.cond.welcome(this);
		ifExpr.a1.welcome(this);
		ifExpr.a2.welcome(this);
		return ifExpr.type=ifExpr.a1.type;
	}

}
